#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "../common/CudaHelper.h"



__global__ void CudaProjectTemplateKernel( float* g_idata, float* g_odata) 
{
	extern __shared__ float shared[];

	const unsigned int tid = threadIdx.x;

	shared[tid] = g_idata[tid];
	__syncthreads();

	shared[tid] = shared[tid] * shared[tid];
	__syncthreads();

	g_odata[tid] = shared[tid];
}



void CudaProjectTemplate( )
{
    unsigned int num_threads = 256;
    unsigned int mem_size = sizeof( float) * num_threads;

	printf("CudaProjectTemplate\n");

    // allocate host memory
    float* h_idata = (float*) malloc( mem_size);
    // initalize the memory
    for( unsigned int i = 0; i < num_threads; ++i) 
    {
        h_idata[i] = (float) i;
    }

    // allocate device memory
    float* d_idata;
    cudaMalloc( (void**) &d_idata, mem_size);
    // copy host memory to device
    cudaMemcpy( d_idata, h_idata, mem_size,
                                cudaMemcpyHostToDevice);

    // allocate device memory for result
    float* d_odata;
    cudaMalloc( (void**) &d_odata, mem_size);

    // setup execution parameters
    dim3  grid( 1, 1, 1);
    dim3  threads( num_threads, 1, 1);

    // execute the kernel
	float timeTaken;
	TIC();
    CudaProjectTemplateKernel<<< grid, threads, mem_size >>>( d_idata, d_odata);
	TOC(timeTaken);

	printf("Time = %f\n",timeTaken);

    // allocate mem for the result on host side
    float* h_odata = (float*) malloc( mem_size);
    // copy result from device to host
    cudaMemcpy( h_odata, d_odata, sizeof(float) * num_threads,
                                cudaMemcpyDeviceToHost);

    // cleanup memory
    free( h_idata);
    free( h_odata);
    cudaFree(d_idata);
    cudaFree(d_odata);
}

int main(){
	printf("main\n");
	CudaProjectTemplate( );
}
