#include <stdio.h>
#include <math.h>
#include "cuda.h"
#include "mex.h"
#include "../common/CudaHelper.h"

#define MAX_THREADS 512
#define MAX_BLOCKS 65536


__global__ void
bandwidthTestKernel( float* g_idata, float* g_odata) 
{
	extern __shared__ float shared[];

	const unsigned int i = (threadIdx.x + blockDim.x*blockIdx.x);

	shared[threadIdx.x] = g_idata[i];
	// to make sure the compiler don't optimize it all away
	//__syncthreads();
	//if(threadIdx.x == 1)
	//	shared[threadIdx.x - 1] += 1.0;
}



int length(const mxArray *vector){
	if( mxGetM(vector) == 1 ){
		return(mxGetN(vector));
	} else if( mxGetN(vector) == 1 ) {
		return(mxGetM(vector));
	} else {
		mexErrMsgTxt("Not a vector");
		return 0;
	}
}


/* Gateway function */
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
	// Get parameters
	//
	// function [transferRate] = bandwidthTest( threads, blocks, reps )

	const mxArray *mx_threads =	prhs[0], 
		          *mx_blocks =	prhs[1], 
				  *mx_reps =	prhs[2];

	// check validity of parameters
	if (nrhs != 3)
		mexErrMsgTxt("Wrong input parameters. bandwidthTest( threads, blocks, reps )");

	double *threads = mxGetPr(mx_threads);
	double *blocks = mxGetPr(mx_blocks);
	double *reps = mxGetPr(mx_reps);

	plhs[0] = mxCreateDoubleMatrix(1, length(mx_threads) * length(mx_blocks) * length(mx_reps), mxREAL);
	double *transferRate = mxGetPr(plhs[0]);


	// allocate memory
	float *g_odata, *g_idata;
	CUDA_SAFE_CALL(cudaMalloc( (void **) &g_idata,sizeof(float)*MAX_THREADS*MAX_BLOCKS));
	CUDA_SAFE_CALL(cudaMalloc( (void **) &g_odata,sizeof(float)*MAX_THREADS*MAX_BLOCKS));

	int i = 0;
	for( int i_r = 0; i_r < length(mx_reps); i_r++){
		for( int i_b = 0; i_b < length(mx_blocks); i_b++){
			for( int i_t = 0; i_t < length(mx_threads); i_t++){
				dim3  dimBlock( (int)threads[i_t] );
				dim3  dimGrid( (int)blocks[i_b] );

				TIC();

				for( int rep = 0; rep < (int)reps[i_r]; rep++)
					bandwidthTestKernel<<<dimGrid,dimBlock,4*dimBlock.x>>>(g_idata, g_odata);

				float timeTaken;
				TOC(timeTaken);

				double bytesTranfered = reps[i_r] * dimGrid.x * dimBlock.x * sizeof( float );
				
				transferRate[i] = bytesTranfered/timeTaken;

				i++;
			}
		}
	}

	CUDA_SAFE_CALL(cudaFree(g_odata));
	CUDA_SAFE_CALL(cudaFree(g_idata));

}

