function sum = beamformApo(RxData, apo, Fs, ac, theta, c0, METHOD)
% function sum = beamformApo(RxData, apo, Fs, ac, theta, c0, METHOD)
%
% RxData: <channels x samples> matrix of data recieved
% apo: <channels x samples> matrix of apodization coefficients
% Fs: sampling frequency.
% ac: Distance from center of beam to center of each transducer element.
% theta: Angle of beam.
% c0: speed of sound
% METHOD: interpolation method (see help interp1)(default is linear)

if nargin == 6
    METHOD = 'linear';
end

N_channels = size(RxData,1);
N_samples = size(RxData,2);

sum = zeros(1,N_samples);
z = linspace(0, 0.5*c0*(N_samples-1)/Fs, N_samples);

t = 2*z/c0;

for i = 1:N_channels
    a = ac(i);
    r = sqrt(z.^2 + a^2 - 2*z*a*sin(theta));
    tau = (z+r)/c0 - t;
    sum = sum + apo(:,i)'.*interp1(t,RxData(i,:),t+tau,METHOD,0);    
end
