#define CUDA_SAFE_CALL( call) do {                                         \
    cudaError err = call;                                                    \
    if( cudaSuccess != err) {                                                \
        fprintf(stderr, "Cuda error in file '%s' in line %i : %s.\n",        \
                __FILE__, __LINE__, cudaGetErrorString( err) );              \
    exit(EXIT_FAILURE);                                                      \
    } } while (0)


#define TIC() \
	cudaEvent_t start, end; \
	CUDA_SAFE_CALL( cudaEventCreate(&start) ); \
	CUDA_SAFE_CALL( cudaEventCreate(&end) ); \
	CUDA_SAFE_CALL( cudaEventRecord(start, 0) )
	

#define TOC(runTime) \
	CUDA_SAFE_CALL( cudaEventRecord(end, 0) ); \
	CUDA_SAFE_CALL( cudaEventSynchronize(end) ); \
 	CUDA_SAFE_CALL( cudaEventElapsedTime(&runTime, start, end) ); \
	runTime /= 1000.0; \
	CUDA_SAFE_CALL( cudaEventDestroy(start) ); \
	CUDA_SAFE_CALL( cudaEventDestroy(end) )
