#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "../common/CudaHelper.h"

#define BLOCK_SIZE 128

texture<float, 1, cudaReadModeNormalizedFloat> tex_short;
texture<float2, 1, cudaReadModeNormalizedFloat> tex_short2;
texture<float4, 1, cudaReadModeNormalizedFloat> tex_short4;

__global__ void read16bitKernel()
{
	__shared__ float shared[BLOCK_SIZE];
	const unsigned int idx = threadIdx.x + blockIdx.x * BLOCK_SIZE*4;
	float a = tex1Dfetch(tex_short, idx);
	a += tex1Dfetch(tex_short, idx+BLOCK_SIZE);
	a += tex1Dfetch(tex_short, idx+BLOCK_SIZE*2);
	shared[threadIdx.x] = a + tex1Dfetch(tex_short, idx+BLOCK_SIZE*3);
}

__global__ void read16bitKernel2()
{
	__shared__ float2 shared[BLOCK_SIZE];
	const unsigned int idx = threadIdx.x + blockIdx.x * BLOCK_SIZE*4;
	float2 a = tex1Dfetch(tex_short2, idx);
	float2 b = tex1Dfetch(tex_short2, idx+BLOCK_SIZE);
	a.x += b.x; a.y += b.y;
	b = tex1Dfetch(tex_short2, idx+BLOCK_SIZE*2);
	a.x += b.x; a.y += b.y;
	b = tex1Dfetch(tex_short2, idx+BLOCK_SIZE*3);
	a.x += b.x; a.y += b.y;
	shared[threadIdx.x] = a;
}

__global__ void read16bitKernel4()
{
	const unsigned int idx = threadIdx.x + blockIdx.x * BLOCK_SIZE*4;
	extern __shared__ float4 shared[];
	float4 a = tex1Dfetch(tex_short4, idx);
	float4 b = tex1Dfetch(tex_short4, idx+BLOCK_SIZE);
	a.x += b.x; a.y += b.y; a.z += b.z; a.w += b.w;
	b = tex1Dfetch(tex_short4, idx+BLOCK_SIZE*2);
	a.x += b.x; a.y += b.y; a.z += b.z; a.w += b.w;
	b = tex1Dfetch(tex_short4, idx+BLOCK_SIZE*3);
	a.x += b.x; a.y += b.y; a.z += b.z; a.w += b.w;
	shared[threadIdx.x] = a;
}

__global__ void copy16bitKernel2(float2 *g_odata)
{
	const unsigned int idx = threadIdx.x + blockIdx.x * BLOCK_SIZE;
	g_odata[idx] = tex1Dfetch(tex_short2, idx);
}



void read16bit( )
{
    size_t num_blocks = 6*32*200;
    size_t mem_size = sizeof(short) * BLOCK_SIZE * num_blocks;




    // allocate host memory
    short* h_idata = (short*) malloc( mem_size);
    // initalize the memory
    for( unsigned int i = 0; i < mem_size/2; ++i) 
    {
        h_idata[i] = (short) i%3456;
    }

    // allocate device memory
    float* d_idata;
    cudaMalloc( (void**) &d_idata, mem_size);

	// copy host memory to device
    cudaMemcpy( d_idata, h_idata, mem_size, cudaMemcpyHostToDevice);

	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc(16, 0, 0, 0, cudaChannelFormatKindSigned);
	CUDA_SAFE_CALL( cudaBindTexture(0, tex_short, d_idata, channelDesc, mem_size) );

	cudaChannelFormatDesc channelDesc2 = cudaCreateChannelDesc(16, 16, 0, 0, cudaChannelFormatKindSigned);
	CUDA_SAFE_CALL( cudaBindTexture(0, tex_short2, d_idata, channelDesc2, mem_size) );

	cudaChannelFormatDesc channelDesc4 = cudaCreateChannelDesc(16, 16, 16, 16, cudaChannelFormatKindSigned);
	CUDA_SAFE_CALL( cudaBindTexture(0, tex_short4, d_idata, channelDesc4, mem_size) );

    // allocate device memory for result
    float* d_odata;
    cudaMalloc( (void**) &d_odata, mem_size*2);

    // setup execution parameters

    // execute the kernel
	{
	    dim3  grid( num_blocks/4, 1, 1);
		dim3  threads( BLOCK_SIZE, 1, 1);
		TIC();

		int reps = 1000;
		for(int i = 0; i < reps; i++)
			read16bitKernel<<< grid, threads, BLOCK_SIZE*4>>>( );

		float timeTaken;
		TOC(timeTaken);

		float data_transferred = (float)reps*mem_size;
		printf("short transferrate = %.2fGB/s\n",data_transferred/(timeTaken*1024*1024*1024));
	}
	{
	    dim3  grid( num_blocks/8, 1, 1);
		dim3  threads( BLOCK_SIZE, 1, 1);
		TIC();

		int reps = 1000;
		for(int i = 0; i < reps; i++)
			read16bitKernel2<<< grid, threads, BLOCK_SIZE*4*2>>>( );

		float timeTaken;
		TOC(timeTaken);

		float data_transferred = (float)reps*mem_size;
		printf("short2 transferrate = %.2fGB/s\n",data_transferred/(timeTaken*1024*1024*1024));
	}
	{
	    dim3  grid( num_blocks/16, 1, 1);
		dim3  threads( BLOCK_SIZE, 1, 1);
		TIC();

		int reps = 1000;
		for(int i = 0; i < reps; i++)
			read16bitKernel4<<< grid, threads, BLOCK_SIZE*4*4>>>( );

		float timeTaken;
		TOC(timeTaken);

		float data_transferred = (float)reps*mem_size;
		printf("short4 transferrate = %.2fGB/s\n",data_transferred/(timeTaken*1024*1024*1024));
	}
	{
	    dim3  grid( num_blocks/2, 1, 1);
		dim3  threads( BLOCK_SIZE, 1, 1);
		TIC();

		int reps = 1;
		for(int i = 0; i < reps; i++)
			copy16bitKernel2<<< grid, threads>>>( (float2*)d_odata );

		float timeTaken;
		TOC(timeTaken);

		float data_transferred = (float)reps*mem_size*3;//1*short + 1*float
		printf("short2 copy transferrate = %.2fGB/s\n",data_transferred/(timeTaken*1024*1024*1024));
	}

    // allocate mem for the result on host side
    float* h_odata = (float*) malloc( mem_size*2);
    // copy result from device to host
    cudaMemcpy( h_odata, d_odata, mem_size*2, cudaMemcpyDeviceToHost);

	// check data
	unsigned int i;
    for( i = 0; i < mem_size/2; ++i) 
    {
		if(h_idata[i] != (short)(h_odata[i]*0x8000)){
			printf("Test failed\n");
			break;
		}
    }

    // cleanup memory
//    free( h_idata);
//    free( h_odata);
    cudaFree(d_idata);
//    cudaFree(d_odata);
}

int main(){
	printf("read16bit\n");
	read16bit( );
}

