#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "../common/CudaHelper.h"



__global__ void sharedMemTestKernelA( float* g_idata, float* g_odata) 
{
	extern __shared__ float shared[];

	// content not important
	const unsigned int tid = threadIdx.x;
	shared[tid] = g_idata[tid];
	__syncthreads();
	g_odata[tid] = shared[tid]*shared[tid];
}

__global__ void sharedMemTestKernelB( float* g_idata, float* g_odata) 
{
	extern __shared__ float shared[];
	__shared__ float mat[100];

	// content not important
	const unsigned int tid = threadIdx.x;
	shared[tid] = g_idata[tid];
	mat[tid%100] = 2;
	__syncthreads();
	g_odata[tid] = shared[tid]*mat[tid%100];
}


__global__ void sharedMemTestKernelC( float* g_odata, float a, float b, float c, float d) 
{
	extern __shared__ float shared[];
	// content not important
	const unsigned int tid = threadIdx.x;
	float x = a;
	float y = b;
	float z = c;
	float w = c+((z<y)?(y-x):(x-y));
	shared[tid] = w;
	__syncthreads();
	g_odata[tid] = shared[tid]*shared[tid];
}

__global__ void sharedMemTestKernelD( ) 
{
	extern __shared__ float shared[];

	// content not important
	const unsigned int tid = threadIdx.x;
	shared[tid] = tid;
}

void checkLaunch(){
	cudaThreadSynchronize();                      
    cudaError_t err = cudaGetLastError();
    if( cudaSuccess != err) {                                          
        fprintf(stderr, "failed : %s.\n",
               cudaGetErrorString( err) );
    } else
		printf("success\n");
}

void sharedMemTest( )
{
    unsigned int num_threads = 256;
    unsigned int mem_size = sizeof( float) * num_threads;

	printf("sharedMemTest\n\n");

    // allocate device memory
    float* d_idata;
    cudaMalloc( (void**) &d_idata, mem_size);

    // allocate device memory for result
    float* d_odata;
    cudaMalloc( (void**) &d_odata, mem_size);

    // setup execution parameters
    dim3  grid( 1, 1, 1);
    dim3  threads( num_threads, 1, 1);

    // execute the kernel
	int sharedSize = 16*1024 - 32;
	printf("Test subject: kernel A\n\n");

	printf("Launching kernel A with shared size = %d...",sharedSize);
    sharedMemTestKernelA<<< grid, threads, sharedSize >>>( d_idata, d_odata);
	checkLaunch();

	sharedSize = 16*1024 - 31 ;
	
	printf("Launching kernel A with shared size = %d...",sharedSize);
    sharedMemTestKernelA<<< grid, threads, sharedSize >>>( d_idata, d_odata);
	checkLaunch();

	printf("\n\nTest subject: kernel B\n");
	printf("  kernel B has a 100 element shared float array\n\n");
	sharedSize = 16*1024 - 32 - 400 ;
	printf("Launching kernel B with shared size = %d...",sharedSize);
    sharedMemTestKernelB<<< grid, threads, sharedSize >>>( d_idata, d_odata);
	checkLaunch();

	sharedSize = 16*1024 - 31 - 400;
	
	printf("Launching kernel B with shared size = %d...",sharedSize);
    sharedMemTestKernelB<<< grid, threads, sharedSize >>>( d_idata, d_odata);
	checkLaunch();

	sharedSize = 16*1024 - 32 - 16 ;
	
	printf("\n\nTest subject: kernel C\n");
	printf("  kernel C has 3 more parameters\n\n");

	printf("Launching kernel C with shared size = %d...",sharedSize);
    sharedMemTestKernelC<<< grid, threads, sharedSize >>>( d_odata, 1, 2, 3, 4 );
	checkLaunch();

	sharedSize = 16*1024 - 32 - 15 ;
	
	printf("Launching kernel C with shared size = %d...",sharedSize);
    sharedMemTestKernelC<<< grid, threads, sharedSize >>>( d_odata, 1, 2, 3, 4 );
	checkLaunch();


    cudaFree(d_idata);
    cudaFree(d_odata);
}

__constant__ char strConst[100];
int main(){

	sharedMemTest( );
}

